
import io
from neurolang.logic import Implication

from ...expressions import Constant, Symbol
from ...frontend import NeurolangPDL
from ..constraints_representation import RightImplication
from ..ontologies_parser import OntologyParser


def test_1():
    owl = '''<?xml version="1.0"?>
    <rdf:RDF xmlns="http://www.w3.org/2002/07/owl#"
        xmlns:owl="http://www.w3.org/2002/07/owl#"
        xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
        xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#">
        <Ontology>
            <versionInfo>0.3.1</versionInfo>
        </Ontology>

        <owl:Class rdf:ID="AdministrativeStaff">
            <rdfs:label>administrative staff worker</rdfs:label>
            <rdfs:subClassOf rdf:resource="#Employee"/>
        </owl:Class>

        <owl:Class rdf:ID="Article">
            <rdfs:label>article</rdfs:label>
            <rdfs:subClassOf rdf:resource="#Publication"/>
        </owl:Class>

        <owl:Class rdf:ID="AssistantProfessor">
            <rdfs:label>assistant professor</rdfs:label>
            <rdfs:subClassOf rdf:resource="#Professor"/>
        </owl:Class>

        <owl:Class rdf:ID="AssociateProfessor">
            <rdfs:label>associate professor</rdfs:label>
            <rdfs:subClassOf rdf:resource="#Professor"/>
        </owl:Class>

        <owl:Class rdf:ID="Book">
            <rdfs:label>book</rdfs:label>
            <rdfs:subClassOf rdf:resource="#Publication"/>
        </owl:Class>
    </rdf:RDF>'''

    x = Symbol.fresh()
    Employee = Symbol('Employee')
    AdministrativeStaff = Symbol('AdministrativeStaff')
    Publication = Symbol('Publication')
    Article = Symbol('Article')
    Professor = Symbol('Professor')
    AssistantProfessor = Symbol('AssistantProfessor')
    AssociateProfessor = Symbol('AssociateProfessor')
    Book = Symbol('Book')

    label = Symbol('rdf-schema:label')

    imp1 = RightImplication(AdministrativeStaff(x), Employee(x))
    imp2 = RightImplication(Article(x), Publication(x))
    imp3 = RightImplication(AssistantProfessor(x), Professor(x))
    imp4 = RightImplication(AssociateProfessor(x), Professor(x))
    imp5 = RightImplication(Book(x), Publication(x))


    imp_label1 = RightImplication(AdministrativeStaff(x), label(x, Constant('administrative staff worker')))
    imp_label2 = RightImplication(AssistantProfessor(x), label(x, Constant('assistant professor')))
    imp_label3 = RightImplication(AssociateProfessor(x), label(x, Constant('associate professor')))
    imp_label4 = RightImplication(Book(x), label(x, Constant('book')))
    imp_label5 = RightImplication(Article(x), label(x, Constant('article')))

    onto = OntologyParser(io.StringIO(owl))
    constraints, _, _ = onto.parse_ontology()

    assert set(constraints.keys()) == set(['Employee', 'Publication', 'Professor', 'rdf-schema:label'])

    AdminConstraint = constraints['Employee']
    assert len(AdminConstraint) == 1 and isinstance(AdminConstraint, set)
    AdminConstraint = next(iter(AdminConstraint))
    assert isinstance(AdminConstraint, RightImplication)
    assert len(AdminConstraint.antecedent.args) == 1
    assert len(AdminConstraint.consequent.args) == 1
    assert AdminConstraint.antecedent.args[0] == AdminConstraint.consequent.args[0]
    assert AdminConstraint.antecedent.functor == imp1.antecedent.functor
    assert AdminConstraint.consequent.functor == imp1.consequent.functor

    PublicationConstraint = constraints['Publication']
    assert len(PublicationConstraint) == 2 and isinstance(PublicationConstraint, set)
    for pc in PublicationConstraint:
        assert isinstance(pc, RightImplication)
        assert len(pc.antecedent.args) == 1
        assert len(pc.consequent.args) == 1
        assert pc.antecedent.args[0] == pc.consequent.args[0]
        if pc.antecedent.functor == Article:
            assert pc.antecedent.functor == imp2.antecedent.functor
            assert pc.consequent.functor == imp2.consequent.functor
        elif pc.antecedent.functor == Book:
            assert pc.antecedent.functor == imp5.antecedent.functor
            assert pc.consequent.functor == imp5.consequent.functor
        else:
            assert False

    ProfessorConstraint = constraints['Professor']
    assert len(ProfessorConstraint) == 2 and isinstance(ProfessorConstraint, set)
    for pc in ProfessorConstraint:
        assert isinstance(pc, RightImplication)
        assert len(pc.antecedent.args) == 1
        assert len(pc.consequent.args) == 1
        assert pc.antecedent.args[0] == pc.consequent.args[0]
        if pc.antecedent.functor == AssistantProfessor:
            assert pc.antecedent.functor == imp3.antecedent.functor
            assert pc.consequent.functor == imp3.consequent.functor
        elif pc.antecedent.functor == AssociateProfessor:
            assert pc.antecedent.functor == imp4.antecedent.functor
            assert pc.consequent.functor == imp4.consequent.functor
        else:
            assert False

    Labels = constraints['rdf-schema:label']
    assert len(Labels) == 5 and isinstance(Labels, set)
    for l in Labels:
        assert isinstance(l, RightImplication)
        assert len(l.antecedent.args) == 1
        assert len(l.consequent.args) == 2
        assert l.antecedent.args[0] == l.consequent.args[0]
        assert isinstance(l.consequent.args[1], Constant)
        assert l.consequent.functor == label
        if l.antecedent.functor == AdministrativeStaff:
            assert l.consequent.args[1] == imp_label1.consequent.args[1]
        elif l.antecedent.functor == AssistantProfessor:
            assert l.consequent.args[1] == imp_label2.consequent.args[1]
        elif l.antecedent.functor == AssociateProfessor:
            assert l.consequent.args[1] == imp_label3.consequent.args[1]
        elif l.antecedent.functor == Book:
            assert l.consequent.args[1] == imp_label4.consequent.args[1]
        elif l.antecedent.functor == Article:
            assert l.consequent.args[1] == imp_label5.consequent.args[1]
        else:
            assert False

def test_2():
    owl = '''<?xml version="1.0"?>
    <rdf:RDF xmlns="http://www.w3.org/2002/07/owl#"
        xmlns:owl="http://www.w3.org/2002/07/owl#"
        xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
        xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#">
        <Ontology>
            <versionInfo>0.3.1</versionInfo>
        </Ontology>

        <owl:Class rdf:ID="ClericalStaff">
            <rdfs:label>clerical staff worker</rdfs:label>
            <rdfs:subClassOf rdf:resource="#AdministrativeStaff"/>
        </owl:Class>

        <owl:Class rdf:ID="College">
            <rdfs:label>school</rdfs:label>
            <rdfs:subClassOf rdf:resource="#Organization"/>
        </owl:Class>

        <owl:Class rdf:ID="ConferencePaper">
            <rdfs:label>conference paper</rdfs:label>
            <rdfs:subClassOf rdf:resource="#Article"/>
        </owl:Class>

        <owl:Class rdf:ID="Course">
            <rdfs:label>teaching course</rdfs:label>
            <rdfs:subClassOf rdf:resource="#Work"/>
        </owl:Class>
    </rdf:RDF>'''

    x = Symbol('x')
    ClericalStaff = Symbol('ClericalStaff')
    AdministrativeStaff = Symbol('AdministrativeStaff')
    College = Symbol('College')
    Organization = Symbol('Organization')
    ConferencePaper = Symbol('ConferencePaper')
    Article = Symbol('Article')
    Course = Symbol('Course')
    Work = Symbol('Work')
    label = Symbol('rdf-schema:label')

    imp1 = RightImplication(ClericalStaff(x), AdministrativeStaff(x))
    imp2 = RightImplication(College(x), Organization(x))
    imp3 = RightImplication(ConferencePaper(x), Article(x))
    imp4 = RightImplication(Course(x), Work(x))

    imp_label1 = RightImplication(ClericalStaff(x), label(x, Constant('clerical staff worker')))
    imp_label2 = RightImplication(College(x), label(x, Constant('school')),)
    imp_label3 = RightImplication(ConferencePaper(x), label(x, Constant('conference paper')))
    imp_label4 = RightImplication(Course(x), label(x, Constant('teaching course')))

    onto = OntologyParser(io.StringIO(owl))
    constraints, _, _ = onto.parse_ontology()

    assert set(constraints.keys()) == set(['AdministrativeStaff', 'Organization', 'Article', 'Work', 'rdf-schema:label'])

    AdminConstraint = constraints['AdministrativeStaff']
    assert len(AdminConstraint) == 1 and isinstance(AdminConstraint, set)
    AdminConstraint = next(iter(AdminConstraint))
    assert isinstance(AdminConstraint, RightImplication)
    assert len(AdminConstraint.antecedent.args) == 1
    assert len(AdminConstraint.consequent.args) == 1
    assert AdminConstraint.antecedent.args[0] == AdminConstraint.consequent.args[0]
    assert AdminConstraint.antecedent.functor == imp1.antecedent.functor
    assert AdminConstraint.consequent.functor == imp1.consequent.functor

    OrganizationConstraint = constraints['Organization']
    assert len(OrganizationConstraint) == 1 and isinstance(OrganizationConstraint, set)
    OrganizationConstraint = next(iter(OrganizationConstraint))
    assert isinstance(OrganizationConstraint, RightImplication)
    assert len(OrganizationConstraint.antecedent.args) == 1
    assert len(OrganizationConstraint.consequent.args) == 1
    assert OrganizationConstraint.antecedent.args[0] == OrganizationConstraint.consequent.args[0]
    assert OrganizationConstraint.antecedent.functor == imp2.antecedent.functor
    assert OrganizationConstraint.consequent.functor == imp2.consequent.functor

    ArticleConstraint = constraints['Article']
    assert len(ArticleConstraint) == 1 and isinstance(ArticleConstraint, set)
    ArticleConstraint = next(iter(ArticleConstraint))
    assert isinstance(ArticleConstraint, RightImplication)
    assert len(ArticleConstraint.antecedent.args) == 1
    assert len(ArticleConstraint.consequent.args) == 1
    assert ArticleConstraint.antecedent.args[0] == ArticleConstraint.consequent.args[0]
    assert ArticleConstraint.antecedent.functor == imp3.antecedent.functor
    assert ArticleConstraint.consequent.functor == imp3.consequent.functor

    WorkConstraint = constraints['Work']
    assert len(WorkConstraint) == 1 and isinstance(WorkConstraint, set)
    WorkConstraint = next(iter(WorkConstraint))
    assert isinstance(WorkConstraint, RightImplication)
    assert len(WorkConstraint.antecedent.args) == 1
    assert len(WorkConstraint.consequent.args) == 1
    assert WorkConstraint.antecedent.args[0] == WorkConstraint.consequent.args[0]
    assert WorkConstraint.antecedent.functor == imp4.antecedent.functor
    assert WorkConstraint.consequent.functor == imp4.consequent.functor


    Labels = constraints['rdf-schema:label']
    assert len(Labels) == 4 and isinstance(Labels, set)
    for l in Labels:
        assert isinstance(l, RightImplication)
        assert len(l.antecedent.args) == 1
        assert len(l.consequent.args) == 2
        assert l.antecedent.args[0] == l.consequent.args[0]
        assert isinstance(l.consequent.args[1], Constant)
        assert l.consequent.functor == label
        if l.antecedent.functor == ClericalStaff:
            assert l.consequent.args[1] == imp_label1.consequent.args[1]
        elif l.antecedent.functor == College:
            assert l.consequent.args[1] == imp_label2.consequent.args[1]
        elif l.antecedent.functor == ConferencePaper:
            assert l.consequent.args[1] == imp_label3.consequent.args[1]
        elif l.antecedent.functor == Course:
            assert l.consequent.args[1] == imp_label4.consequent.args[1]
        else:
            assert False


def test_3():
    owl = '''<?xml version="1.0"?>
    <rdf:RDF xmlns="http://www.w3.org/2002/07/owl#"
        xmlns:owl="http://www.w3.org/2002/07/owl#"
        xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
        xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#">
        <Ontology>
            <versionInfo>0.3.1</versionInfo>
        </Ontology>

        <owl:Class rdf:ID="Chair">
            <rdfs:label>chair</rdfs:label>
            <rdfs:subClassOf>
                <owl:Class>
                    <owl:intersectionOf rdf:parseType="Collection">
                        <owl:Class rdf:about="#Person" />
                        <owl:Restriction>
                            <owl:onProperty rdf:resource="#headOf" />
                            <owl:someValuesFrom>
                                <owl:Class rdf:about="#Department" />
                            </owl:someValuesFrom>
                        </owl:Restriction>
                    </owl:intersectionOf>
                </owl:Class>
            </rdfs:subClassOf>
            <rdfs:subClassOf rdf:resource="#Professor" />
        </owl:Class>
    </rdf:RDF>'''

    Professor = Symbol('Professor')
    Chair = Symbol('Chair')
    headOf = Symbol('headOf')
    Department = Symbol('Department')
    Person = Symbol('Person')

    label = Symbol('rdf-schema:label')

    onto = OntologyParser(io.StringIO(owl))
    constraints, _, _ = onto.parse_ontology()

    for c in constraints:
        if c.startswith('fresh'):
            support_rule = next(iter(constraints[c]))
            supportChair = support_rule.consequent.functor
            x = support_rule.consequent.args[0]
            y = support_rule.consequent.args[1]
            break

    assert set(constraints.keys()) == set(['Person', 'headOf', 'Department', 'Professor', 'rdf-schema:label', c])

    imp1 = RightImplication(supportChair(x, y), Person(x))
    imp2 = RightImplication(Chair(x), supportChair(x, y))
    imp3 = RightImplication(supportChair(x, y), headOf(x, y))
    imp4 = RightImplication(supportChair(x, y), Department(y))
    imp5 = RightImplication(Chair(x), Professor(x),)

    imp_label = RightImplication(Chair(x), label(x, Constant('chair')))

    PersonConstraint = constraints['Person']
    assert len(PersonConstraint) == 1 and isinstance(PersonConstraint, set)
    PersonConstraint = next(iter(PersonConstraint))
    assert isinstance(PersonConstraint, RightImplication)
    assert len(PersonConstraint.antecedent.args) == 2
    assert len(PersonConstraint.consequent.args) == 1
    assert PersonConstraint.antecedent.args[0] == PersonConstraint.consequent.args[0]
    assert PersonConstraint.antecedent.functor == imp1.antecedent.functor
    assert PersonConstraint.consequent.functor == imp1.consequent.functor

    SupportConstraint = constraints[c]
    assert len(SupportConstraint) == 1 and isinstance(SupportConstraint, set)
    SupportConstraint = next(iter(SupportConstraint))
    assert isinstance(SupportConstraint, RightImplication)
    assert len(SupportConstraint.antecedent.args) == 1
    assert len(SupportConstraint.consequent.args) == 2
    assert SupportConstraint.antecedent.args[0] == SupportConstraint.consequent.args[0]
    assert SupportConstraint.antecedent.functor == imp2.antecedent.functor
    assert SupportConstraint.consequent.functor == imp2.consequent.functor

    headOfConstraint = constraints['headOf']
    assert len(headOfConstraint) == 1 and isinstance(headOfConstraint, set)
    headOfConstraint = next(iter(headOfConstraint))
    assert isinstance(headOfConstraint, RightImplication)
    assert len(headOfConstraint.antecedent.args) == 2
    assert len(headOfConstraint.consequent.args) == 2
    assert headOfConstraint.antecedent.args == headOfConstraint.consequent.args
    assert headOfConstraint.antecedent.functor == imp3.antecedent.functor
    assert headOfConstraint.consequent.functor == imp3.consequent.functor

    DepartmentConstraint = constraints['Department']
    assert len(DepartmentConstraint) == 1 and isinstance(DepartmentConstraint, set)
    DepartmentConstraint = next(iter(DepartmentConstraint))
    assert isinstance(DepartmentConstraint, RightImplication)
    assert len(DepartmentConstraint.antecedent.args) == 2
    assert len(DepartmentConstraint.consequent.args) == 1
    assert DepartmentConstraint.antecedent.args[1] == DepartmentConstraint.consequent.args[0]
    assert DepartmentConstraint.antecedent.functor == imp4.antecedent.functor
    assert DepartmentConstraint.consequent.functor == imp4.consequent.functor

    ProfessorConstraint = constraints['Professor']
    assert len(ProfessorConstraint) == 1 and isinstance(ProfessorConstraint, set)
    ProfessorConstraint = next(iter(ProfessorConstraint))
    assert isinstance(ProfessorConstraint, RightImplication)
    assert len(ProfessorConstraint.antecedent.args) == 1
    assert len(ProfessorConstraint.consequent.args) == 1
    assert ProfessorConstraint.antecedent.args[0] == ProfessorConstraint.consequent.args[0]
    assert ProfessorConstraint.antecedent.functor == imp5.antecedent.functor
    assert ProfessorConstraint.consequent.functor == imp5.consequent.functor

    Labels = constraints['rdf-schema:label']
    assert len(Labels) == 1 and isinstance(Labels, set)
    l = Labels.pop()
    assert isinstance(l, RightImplication)
    assert len(l.antecedent.args) == 1
    assert len(l.consequent.args) == 2
    assert l.antecedent.args[0] == l.consequent.args[0]
    assert isinstance(l.consequent.args[1], Constant)
    assert l.consequent.functor == label
    if l.antecedent.functor == Chair:
        assert l.consequent.args[1] == imp_label.consequent.args[1]
    else:
        assert False



def test_open_world_example():
    owl = '''<?xml version="1.0"?>
    <rdf:RDF xmlns="http://www.w3.org/2002/07/owl#"
        xmlns:owl="http://www.w3.org/2002/07/owl#"
        xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
        xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#">
        <Ontology>
            <versionInfo>0.3.1</versionInfo>
        </Ontology>

        <owl:Class rdf:ID="Dean">
            <rdfs:label>dean</rdfs:label>
            <rdfs:subClassOf>
                <owl:Class>
                    <owl:intersectionOf rdf:parseType="Collection">
                        <owl:Class rdf:about="#Person" />
                        <owl:Restriction>
                            <owl:onProperty rdf:resource="#headOf" />
                            <owl:someValuesFrom>
                                <owl:Class rdf:about="#College" />
                            </owl:someValuesFrom>
                        </owl:Restriction>
                    </owl:intersectionOf>
                </owl:Class>
            </rdfs:subClassOf>
            <rdfs:subClassOf rdf:resource="#Professor" />
        </owl:Class>

        <owl:Class rdf:ID="Chair">
            <rdfs:label>chair</rdfs:label>
            <rdfs:subClassOf>
                <owl:Class>
                    <owl:intersectionOf rdf:parseType="Collection">
                        <owl:Class rdf:about="#Person" />
                        <owl:Restriction>
                            <owl:onProperty rdf:resource="#headOf" />
                            <owl:someValuesFrom>
                                <owl:Class rdf:about="#Department" />
                            </owl:someValuesFrom>
                        </owl:Restriction>
                    </owl:intersectionOf>
                </owl:Class>
            </rdfs:subClassOf>
            <rdfs:subClassOf rdf:resource="#Professor" />
        </owl:Class>
    </rdf:RDF>'''

    nl = NeurolangPDL()
    nl.load_ontology(io.StringIO(owl))

    nl.add_tuple_set([('Juan',), ('Manuel',)], name='Dean')
    nl.add_tuple_set([('Miguel',), ('Alberto',)], name='Chair')
    nl.add_tuple_set([('College A',), ('College B',)], name='College')
    nl.add_tuple_set([('Department A',), ('Department B',)], name='Department')

    with nl.scope as e:
        e.answer[e.a] = (
            e.headOf[e.a, e.b] &
            e.College[e.b]
        )

        f_term = nl.solve_all()

    res = f_term['answer'].as_pandas_dataframe().values
    assert (res == [['Juan'], ['Manuel']]).all()


def test_retrieve_property():
    owl = '''<?xml version="1.0"?>
    <rdf:RDF xmlns="http://www.w3.org/2002/07/owl#"
        xmlns:owl="http://www.w3.org/2002/07/owl#"
        xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
        xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#">
        <Ontology>
            <versionInfo>0.3.1</versionInfo>
        </Ontology>

        <owl:Class rdf:ID="Dean">
            <rdfs:label>dean</rdfs:label>
            <rdfs:subClassOf>
                <owl:Class>
                    <owl:intersectionOf rdf:parseType="Collection">
                        <owl:Class rdf:about="#Person" />
                        <owl:Restriction>
                            <owl:onProperty rdf:resource="#headOf" />
                            <owl:someValuesFrom>
                                <owl:Class rdf:about="#College" />
                            </owl:someValuesFrom>
                        </owl:Restriction>
                    </owl:intersectionOf>
                </owl:Class>
            </rdfs:subClassOf>
            <rdfs:subClassOf rdf:resource="#Professor" />
        </owl:Class>

        <owl:Class rdf:ID="Chair">
            <rdfs:label>chair</rdfs:label>
            <rdfs:subClassOf>
                <owl:Class>
                    <owl:intersectionOf rdf:parseType="Collection">
                        <owl:Class rdf:about="#Person" />
                        <owl:Restriction>
                            <owl:onProperty rdf:resource="#headOf" />
                            <owl:someValuesFrom>
                                <owl:Class rdf:about="#Department" />
                            </owl:someValuesFrom>
                        </owl:Restriction>
                    </owl:intersectionOf>
                </owl:Class>
            </rdfs:subClassOf>
            <rdfs:subClassOf rdf:resource="#Professor" />
        </owl:Class>
    </rdf:RDF>'''

    nl = NeurolangPDL()
    nl.load_ontology(io.StringIO(owl))

    nl.add_tuple_set([('Juan',), ('Manuel',)], name='Dean')
    nl.add_tuple_set([('Miguel',), ('Alberto',)], name='Chair')
    nl.add_tuple_set([('College A',), ('College B',)], name='College')
    nl.add_tuple_set([('Department A',), ('Department B',)], name='Department')

    label = nl.new_symbol(name='rdf-schema:label')

    with nl.scope as e:
        e.answer[e.a] = (
            label[e.a, 'dean']
        )

        f_term = nl.solve_all()

    res = f_term['answer'].as_pandas_dataframe().values
    assert (res == [['Juan'], ['Manuel']]).all()

def test_retrieve_subclass():
    owl = '''<?xml version="1.0"?>
    <rdf:RDF xmlns="http://www.w3.org/2002/07/owl#"
        xmlns:owl="http://www.w3.org/2002/07/owl#"
        xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
        xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#">
        <Ontology>
            <versionInfo>0.3.1</versionInfo>
        </Ontology>

        <owl:Class rdf:ID="Dean">
            <rdfs:label>dean</rdfs:label>
            <rdfs:subClassOf>
                <owl:Class>
                    <owl:intersectionOf rdf:parseType="Collection">
                        <owl:Class rdf:about="#Person" />
                        <owl:Restriction>
                            <owl:onProperty rdf:resource="#headOf" />
                            <owl:someValuesFrom>
                                <owl:Class rdf:about="#College" />
                            </owl:someValuesFrom>
                        </owl:Restriction>
                    </owl:intersectionOf>
                </owl:Class>
            </rdfs:subClassOf>
            <rdfs:subClassOf rdf:resource="#Professor" />
        </owl:Class>

        <owl:Class rdf:ID="Chair">
            <rdfs:label>chair</rdfs:label>
            <rdfs:subClassOf>
                <owl:Class>
                    <owl:intersectionOf rdf:parseType="Collection">
                        <owl:Class rdf:about="#Person" />
                        <owl:Restriction>
                            <owl:onProperty rdf:resource="#headOf" />
                            <owl:someValuesFrom>
                                <owl:Class rdf:about="#Department" />
                            </owl:someValuesFrom>
                        </owl:Restriction>
                    </owl:intersectionOf>
                </owl:Class>
            </rdfs:subClassOf>
            <rdfs:subClassOf rdf:resource="#Professor" />
        </owl:Class>
    </rdf:RDF>'''

    nl = NeurolangPDL()
    nl.load_ontology(io.StringIO(owl))

    nl.add_tuple_set([('Juan',), ('Manuel',)], name='Dean')
    nl.add_tuple_set([('Miguel',), ('Alberto',)], name='Chair')
    nl.add_tuple_set([('College A',), ('College B',)], name='College')
    nl.add_tuple_set([('Department A',), ('Department B',)], name='Department')

    with nl.scope as e:
        e.answer[e.a] = (
            e.Professor[e.a]
        )

        f_term = nl.solve_all()

    res = f_term['answer'].as_pandas_dataframe().values
    assert ['Miguel'] in res
    assert ['Alberto'] in res
    assert ['Juan'] in res
    assert ['Manuel'] in res
    assert len(res) == 4


def test_knowledge_subclassof():
    owl = '''<?xml version="1.0"?>
    <rdf:RDF xmlns="http://www.w3.org/2002/07/owl#"
        xmlns:owl="http://www.w3.org/2002/07/owl#"
        xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
        xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#">
        <Ontology>
            <versionInfo>0.3.1</versionInfo>
        </Ontology>

        <owl:Class rdf:ID="Chair">
            <rdfs:label>chair</rdfs:label>
            <rdfs:subClassOf>
                <owl:Class>
                    <owl:intersectionOf rdf:parseType="Collection">
                        <owl:Class rdf:about="#Person" />
                        <owl:Restriction>
                            <owl:onProperty rdf:resource="#headOf" />
                            <owl:someValuesFrom>
                                <owl:Class rdf:about="#Department" />
                            </owl:someValuesFrom>
                        </owl:Restriction>
                    </owl:intersectionOf>
                </owl:Class>
            </rdfs:subClassOf>
            <rdfs:subClassOf rdf:resource="#Professor" />
        </owl:Class>
    </rdf:RDF>'''

    nl = NeurolangPDL()
    nl.load_ontology(io.StringIO(owl))

    subClassOf = nl.new_symbol(name='neurolang:subClassOf')
    with nl.scope as e:
        e.answer[e.a] = (
            subClassOf[e.a, 'Professor']
        )

        f_term = nl.query((e.a,), e.answer(e.a))

    res = f_term.as_pandas_dataframe().values
    assert (res == [['Chair']]).all()

def test_knowledge_property():
    owl = '''<?xml version="1.0"?>
    <rdf:RDF xmlns="http://www.w3.org/2002/07/owl#"
        xmlns:owl="http://www.w3.org/2002/07/owl#"
        xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
        xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#">
        <Ontology>
            <versionInfo>0.3.1</versionInfo>
        </Ontology>

        <owl:Class rdf:ID="Chair">
            <rdfs:label>chair</rdfs:label>
            <rdfs:subClassOf>
                <owl:Class>
                    <owl:intersectionOf rdf:parseType="Collection">
                        <owl:Class rdf:about="#Person" />
                        <owl:Restriction>
                            <owl:onProperty rdf:resource="#headOf" />
                            <owl:someValuesFrom>
                                <owl:Class rdf:about="#Department" />
                            </owl:someValuesFrom>
                        </owl:Restriction>
                    </owl:intersectionOf>
                </owl:Class>
            </rdfs:subClassOf>
            <rdfs:subClassOf rdf:resource="#Professor" />
        </owl:Class>
    </rdf:RDF>'''

    nl = NeurolangPDL()
    nl.load_ontology(io.StringIO(owl))

    label = nl.new_symbol(name='neurolang:label')
    with nl.scope as e:
        e.answer[e.a] = (
            label[e.a, 'chair']
        )

        f_term = nl.query((e.a,), e.answer[e.a])

    res = f_term.as_pandas_dataframe().values
    assert (res == [['Chair']]).all()



def test_entity_rules():
    owl = '''<?xml version="1.0"?>
    <rdf:RDF xmlns="http://www.w3.org/2002/07/owl#"
        xmlns:owl="http://www.w3.org/2002/07/owl#"
        xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
        xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#">
        <Ontology>
            <versionInfo>0.3.1</versionInfo>
        </Ontology>
        <owl:Class rdf:ID="Chair">
            <rdfs:label>chair</rdfs:label>
            <rdfs:subClassOf>
                <owl:Class>
                    <owl:intersectionOf rdf:parseType="Collection">
                        <owl:Class rdf:about="#Person" />
                        <owl:Restriction>
                            <owl:onProperty rdf:resource="#headOf" />
                            <owl:someValuesFrom>
                                <owl:Class rdf:about="#Department" />
                            </owl:someValuesFrom>
                        </owl:Restriction>
                    </owl:intersectionOf>
                </owl:Class>
            </rdfs:subClassOf>
            <rdfs:subClassOf rdf:resource="#Professor" />
        </owl:Class>
    </rdf:RDF>'''

    entity = Symbol('Entity')
    onto = OntologyParser(io.StringIO(owl), connector_symbol=entity)
    _, _, entity_rules = onto.parse_ontology()

    chair = Symbol('Chair')
    keys = entity_rules.keys()
    assert len(keys) == 1
    assert chair in keys

    rule = entity_rules[chair]
    assert len(rule) == 1
    rule = rule.pop()
    assert isinstance(rule, Implication)
    assert rule.consequent.functor == chair
    assert len(rule.consequent.args) == 1 and rule.consequent.args[0].is_fresh

    fresh_var = rule.consequent.args[0]
    assert rule == Implication(chair(fresh_var), entity(fresh_var, Constant('chair')))


def test_entity_rules_inclusion():
    owl = '''<?xml version="1.0"?>
    <rdf:RDF xmlns="http://www.w3.org/2002/07/owl#"
        xmlns:owl="http://www.w3.org/2002/07/owl#"
        xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
        xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#">
        <Ontology>
            <versionInfo>0.3.1</versionInfo>
        </Ontology>
        <owl:Class rdf:ID="Chair">
            <rdfs:label>chair</rdfs:label>
            <rdfs:subClassOf>
                <owl:Class>
                    <owl:intersectionOf rdf:parseType="Collection">
                        <owl:Class rdf:about="#Person" />
                        <owl:Restriction>
                            <owl:onProperty rdf:resource="#headOf" />
                            <owl:someValuesFrom>
                                <owl:Class rdf:about="#Department" />
                            </owl:someValuesFrom>
                        </owl:Restriction>
                    </owl:intersectionOf>
                </owl:Class>
            </rdfs:subClassOf>
            <rdfs:subClassOf rdf:resource="#Professor" />
        </owl:Class>
    </rdf:RDF>'''

    nl = NeurolangPDL()

    entity = nl.load_ontology(io.StringIO(owl), connector_symbol_name='Entity')

    chair = Symbol('Chair')
    assert chair in nl.symbol_table

    assert len(nl.symbol_table[chair].formulas) == 1
    rule = nl.symbol_table[chair].formulas[0]

    assert isinstance(rule, Implication)
    assert rule.consequent.functor == chair
    assert len(rule.consequent.args) == 1 and rule.consequent.args[0].is_fresh

    fresh_var = rule.consequent.args[0]
    assert rule == Implication(chair(fresh_var), entity(fresh_var, Constant('chair')))




from neurolang.utils.orderedset import OrderedSet
from neurolang.logic.expression_processing import extract_logic_free_variables
from neurolang.logic.transformations import (
    CheckConjunctiveQueryWithNegation, GuaranteeConjunction, ExtractConjunctiveQueryWithNegation,
    PushExistentialsDown, RemoveExistentialPredicates, RemoveTrivialOperations
)
from neurolang.probabilistic.dalvi_suciu_lift import convert_ucq_to_ccq
from neurolang.probabilistic.transforms import convert_rule_to_ucq
from neurolang.logic import (
    Conjunction, Disjunction, ExistentialPredicate, Implication
)
from neurolang.expressions import Symbol

ECQ = ExtractConjunctiveQueryWithNegation()
CPC = CheckConjunctiveQueryWithNegation()
REP = RemoveExistentialPredicates()
PED = PushExistentialsDown()
GC = GuaranteeConjunction()

# Tests for RemoveExistentialPredicate, start here
def test_remove_existential():
    S = Symbol('S')
    x = Symbol('x')
    y = Symbol('y')

    rule = ExistentialPredicate(y, S(x, y))
    rule = REP.walk(rule)

    expected = S(x, y)
    assert rule == expected

def test_remove_inner_existential():
    S = Symbol('S')
    x = Symbol('x')
    y = Symbol('y')

    rule = Conjunction((
        ExistentialPredicate(y, S(x, y)),
    ))

    rule = REP.walk(rule)

    expected = Conjunction((
        S(x, y),
    ))
    assert rule == expected

def test_remove_nested_existentials():
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x1 = Symbol('x1')
    x2 = Symbol('x2')
    y1 = Symbol('y1')
    y2 = Symbol('y2')
    z = Symbol('z')

    rule = Conjunction((
        ExistentialPredicate(y2,
            ExistentialPredicate(x2,
                Disjunction((
                    T(z, x2), S(x2, y2)
                ))
            )
        ),
        ExistentialPredicate(y1,
            ExistentialPredicate(x1,
                Conjunction((
                    R(z, x1), S(x1, y1)
                ))
            )
        )
    ))

    rule = REP.walk(rule)

    expected = Conjunction((
        Disjunction((
            T(z, x2), S(x2, y2)
        )),
        Conjunction((
            R(z, x1), S(x1, y1)
        )),
    ))
    assert rule == expected

# Tests for CheckPureConjunction, start here
def test_pure_conjunction():
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x1 = Symbol('x1')
    x2 = Symbol('x2')
    y1 = Symbol('y1')
    y2 = Symbol('y2')
    z = Symbol('z')

    rule = Conjunction((
         R(z, x1), S(x1, y1), T(z, x2), S(x2, y2)
    ))

    is_pure = CPC.walk(rule)
    assert is_pure

def test_pure_conjunction_with_existentials():
    S = Symbol('S')
    x = Symbol('x')
    y = Symbol('y')

    rule = Conjunction((
        ExistentialPredicate(y, S(x, y)),
    ))

    is_pure = CPC.walk(rule)
    assert is_pure

def test_nested_conjunctions():
    P = Symbol('P')
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x = Symbol('x')
    y = Symbol('y')

    rule = Conjunction((
        Conjunction((
             P(x), S(x), R(y)
        )),
        Conjunction((
            T(x), S(y)
        )),
        R(y)
    ))

    is_pure = CPC.walk(rule)
    assert not is_pure

def test_nested_disjunction():
    P = Symbol('P')
    R = Symbol('R')
    S = Symbol('S')
    x = Symbol('x')
    y = Symbol('y')

    rule = Disjunction((
        Conjunction((
             P(x), S(x), R(y)
        )),
    ))

    is_pure = CPC.walk(rule)
    assert not is_pure

def test_expression():
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x1 = Symbol('x1')
    x2 = Symbol('x2')
    y1 = Symbol('y1')
    y2 = Symbol('y2')
    z = Symbol('z')

    rule = Conjunction((
        ExistentialPredicate(y2,
            ExistentialPredicate(x2,
                Conjunction((
                    T(z, x2), S(x2, y2)
                ))
            )
        ),
        ExistentialPredicate(y1,
            ExistentialPredicate(x1,
                Conjunction((
                    R(z, x1), S(x1, y1)
                ))
            )
        )
    ))

    is_pure = CPC.walk(rule)
    assert not is_pure

# Tests for IdentifyPureConjunctions, start here
def test_identify_pure_conjunction():
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x1 = Symbol('x1')
    x2 = Symbol('x2')
    y1 = Symbol('y1')
    y2 = Symbol('y2')
    z = Symbol('z')

    rule = Conjunction((
         R(z, x1), S(x1, y1), T(z, x2), S(x2, y2)
    ))

    expected = OrderedSet([rule])

    conjunctions = ECQ.walk(rule)
    assert conjunctions == expected


def test_identify_existential_conjunctions():
    S = Symbol('S')
    x = Symbol('x')
    y = Symbol('y')

    rule = Conjunction((
        ExistentialPredicate(y, S(x, y)),
    ))

    expected = OrderedSet([rule])

    conjunctions = ECQ.walk(rule)
    assert conjunctions == expected


def test_basic_nested_conjunctions():
    P = Symbol('P')
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x = Symbol('x')
    y = Symbol('y')

    rule = Conjunction((
        Conjunction((
             P(x), S(x), R(y)
        )),
        Conjunction((
            T(x), S(y)
        )),
        R(y)
    ))

    expected = OrderedSet([rule.formulas[0], rule.formulas[1]])

    conjunctions = ECQ.walk(rule)
    assert conjunctions == expected

def test_basic_inner_disjunctions():
    P = Symbol('P')
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x = Symbol('x')
    y = Symbol('y')

    rule = Conjunction((
        Disjunction((
             P(x), S(x), R(y)
        )),
        Disjunction((
            T(x), S(y)
        )),
        R(y)
    ))

    expected = OrderedSet([])

    conjunctions = ECQ.walk(rule)
    assert conjunctions == expected

def test_basic_nested_mixed():
    P = Symbol('P')
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x = Symbol('x')
    y = Symbol('y')

    rule = Conjunction((
        Conjunction((
             P(x), S(x), R(y)
        )),
        Disjunction((
            T(x), S(y)
        )),
        R(y)
    ))

    expected = OrderedSet([rule.formulas[0]])

    conjunctions = ECQ.walk(rule)
    assert conjunctions == expected


def test_basic_nested_mixed_2():
    P = Symbol('P')
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x = Symbol('x')
    y = Symbol('y')

    rule = Conjunction((
        P(x),
        Disjunction((
            T(x), S(y)
        )),
        R(y)
    ))

    expected = OrderedSet([])

    conjunctions = ECQ.walk(rule)
    assert conjunctions == expected


def test_nested_existentials():
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x1 = Symbol('x1')
    x2 = Symbol('x2')
    y1 = Symbol('y1')
    y2 = Symbol('y2')
    z = Symbol('z')

    rule = Conjunction((
        ExistentialPredicate(y2,
            ExistentialPredicate(x2,
                Conjunction((
                    T(z, x2), S(x2, y2)
                ))
            )
        ),
        ExistentialPredicate(y1,
            ExistentialPredicate(x1,
                Conjunction((
                    R(z, x1), S(x1, y1)
                ))
            )
        )
    ))

    expected = OrderedSet([rule.formulas[0], rule.formulas[1]])

    conjunctions = ECQ.walk(rule)
    assert conjunctions == expected


# Tests for CCQ transformation, start here

def test_ccq_transformation_example_2_12():
    Q = Symbol('Q')
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x1 = Symbol('x1')
    x2 = Symbol('x2')
    x3 = Symbol('x3')
    y1 = Symbol('y1')
    y2 = Symbol('y2')
    y3 = Symbol('y3')
    z = Symbol('z')

    rule = Implication(Q(z), Disjunction((
        Conjunction((
            R(x1), S(x1, y1),
        )),
        Conjunction((
            T(y2), S(x2, y2),
        )),
        Conjunction((
            R(x3), T(y3),
        )),
    )))

    rule = convert_rule_to_ucq(rule)
    rule = convert_ucq_to_ccq(rule)

    c0 = ExistentialPredicate(x1,
        Conjunction((
            ExistentialPredicate(
                y1, S(x1, y1),
            ),
            R(x1),
        ))
    )
    c1 = ExistentialPredicate(y2,
        Conjunction((
            ExistentialPredicate(
                x2, S(x2, y2),
            ),
            T(y2),
        ))
    )
    c2 = ExistentialPredicate(x3, R(x3),)
    c3 = ExistentialPredicate(y3, T(y3),)

    expected = Conjunction((
        Disjunction((
            c0, c1, c3
        )),
        Disjunction((
            c0, c1, c2
        )),
    ))

    assert rule == expected


def test_ccq_no_transformation_conjunction():

    Q = Symbol('Q')
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x1 = Symbol('x1')
    x2 = Symbol('x2')
    y1 = Symbol('y1')
    y2 = Symbol('y2')
    z = Symbol('z')

    rule = Implication(Q(z), Conjunction((
         R(z, x1), S(x1, y1), T(z, x2), S(x2, y2)
    )))

    rule_ucq = convert_rule_to_ucq(rule)
    rule_ccq = convert_ucq_to_ccq(rule_ucq)

    RTO = RemoveTrivialOperations()
    implication = RTO.walk(rule)
    consequent, antecedent = implication.unapply()
    head_vars = set(consequent.args)
    existential_vars = set(
        extract_logic_free_variables(antecedent) -
        set(head_vars)
    )

    for a in existential_vars:
        antecedent = ExistentialPredicate(a, antecedent)
    expected = RTO.walk(PED.walk(antecedent))

    assert isinstance(rule_ccq, Conjunction)
    assert isinstance(expected, Conjunction)
    assert len(expected.formulas) == len(rule_ccq.formulas)
    # The formulas are not sorted, but they are always the same
    assert (
        (rule_ccq.formulas[0] == expected.formulas[0] and
        rule_ccq.formulas[1] == expected.formulas[1]) or

        (rule_ccq.formulas[0] == expected.formulas[1] and
        rule_ccq.formulas[1] == expected.formulas[0])
    )



# -*- coding: utf-8 -*-
r"""
NeuroLang Example based Implementing a NeuroSynth Query
====================================================

"""

# %%
import warnings

warnings.filterwarnings("ignore")

from pathlib import Path
from typing import Iterable

import nibabel
import nilearn.datasets
import nilearn.image
import nilearn.plotting
import numpy as np
import pandas as pd
from neurolang import ExplicitVBROverlay, NeurolangPDL
from neurolang.frontend.neurosynth_utils import get_ns_mni_peaks_reported

###############################################################################
# Data preparation
# ----------------

data_dir = Path.home() / "neurolang_data"

###############################################################################
# Load the MNI atlas and resample it to 4mm voxels

mni_t1 = nibabel.load(
    nilearn.datasets.fetch_icbm152_2009(data_dir=str(data_dir / "icbm"))["t1"]
)
mni_t1_4mm = nilearn.image.resample_img(mni_t1, np.eye(3) * 4)

###############################################################################
# Probabilistic Logic Programming in NeuroLang
# --------------------------------------------

nl = NeurolangPDL()


###############################################################################
# Adding new aggregation function to build a region overlay


@nl.add_symbol
def agg_create_region_overlay(
    i: Iterable, j: Iterable, k: Iterable, p: Iterable
) -> ExplicitVBROverlay:
    mni_coords = np.c_[i, j, k]
    return ExplicitVBROverlay(
        mni_coords, mni_t1_4mm.affine, p, image_dim=mni_t1_4mm.shape
    )


###############################################################################
# Load the NeuroSynth database

peak_data = get_ns_mni_peaks_reported(data_dir)
ijk_positions = np.round(
    nibabel.affines.apply_affine(
        np.linalg.inv(mni_t1_4mm.affine),
        peak_data[["x", "y", "z"]].values.astype(float),
    )
).astype(int)
peak_data["i"] = ijk_positions[:, 0]
peak_data["j"] = ijk_positions[:, 1]
peak_data["k"] = ijk_positions[:, 2]
peak_data = peak_data[["i", "j", "k", "id"]]

nl.add_tuple_set(peak_data, name="PeakReported")
study_ids = nl.load_neurosynth_study_ids(data_dir, "Study")
nl.add_uniform_probabilistic_choice_over_set(
    study_ids.value, name="SelectedStudy"
)
nl.load_neurosynth_term_study_associations(
    data_dir, "TermInStudyTFIDF", tfidf_threshold=1e-3
)


# %%
###############################################################################
# Probabilistic program and querying

with nl.scope as e:
    e.Activation[e.i, e.j, e.k] = e.PeakReported(
        e.i, e.j, e.k, e.s
    ) & e.SelectedStudy(e.s)
    e.TermAssociation[e.t] = e.SelectedStudy(e.s) & e.TermInStudyTFIDF(
        e.s, e.t, ...
    )
    e.ActivationGivenTerm[e.i, e.j, e.k, e.PROB[e.i, e.j, e.k]] = e.Activation(
        e.i, e.j, e.k
    ) // e.TermAssociation("auditory")
    e.ActivationGivenTermImage[
        agg_create_region_overlay(e.i, e.j, e.k, e.p)
    ] = e.ActivationGivenTerm(e.i, e.j, e.k, e.p)

    img_query = nl.query((e.x,), e.ActivationGivenTermImage(e.x))


# %%
###############################################################################
# Plotting results
# --------------------------------------------

result_image = img_query.fetch_one()[0].spatial_image()
img = result_image.get_fdata()
plot = nilearn.plotting.plot_stat_map(
    result_image, threshold=np.percentile(img[img > 0], 95)
)
nilearn.plotting.show()



import configparser
import logging
import os
import sys
from importlib import reload

LOG = logging.getLogger(__name__)


class NeurolangConfigParser(configparser.ConfigParser):
    def set_query_backend(self, backend):
        """
        Convenience method to set the backend used by Neurolang.
        """
        valid_backends = {"pandas", "dask"}
        if backend not in valid_backends:
            raise ValueError(
                f"The query backend option should be one of {valid_backends}"
            )
        LOG.info(f"Setting new query backend : {backend}")
        self["RAS"]["backend"] = backend
        self.switch_backend()

    def switch_backend(self):
        """
        Changing the backend value in the config object is not enough to
        switch the backends since this config value is evaluated at import
        time in some modules to figure out which classes to import.
        So we need to reimport these modules after changing the backend value
        in the config.
        """
        modules = [
            "neurolang.utils.relational_algebra_set",
            "neurolang.utils",
            "neurolang.probabilistic.weighted_model_counting",
        ]
        for module in modules:
            module = sys.modules.get(module)
            if module is not None:
                reload(module)
                LOG.debug(f"Reloading module : {module}")

    def enable_expression_type_printing(self):
        self.set("DEFAULT", "expressionTypePrinting", "True")

    def disable_expression_type_printing(self):
        self.set("DEFAULT", "expressionTypePrinting", "False")

    def expression_type_printing(self):
        return self["DEFAULT"].get("expressionTypePrinting", "True") == "True"

    def switch_expression_type_printing(self):
        old_value = self.expression_type_printing()
        self["DEFAULT"]["expressionTypePrinting"] = str(not old_value)
        return self["DEFAULT"]["expressionTypePrinting"] == "True"

    def set_structural_knowledge_namespace(self, name):
        self.set("DEFAULT", "structural_knowledge_namespace", name)

    def get_structural_knowledge_namespace(self):
        return self["DEFAULT"].get("structural_knowledge_namespace", "neurolang:")

    def get_probabilistic_solver_check_unate(self):
        return self.getboolean("PROBABILISTIC_SOLVER", "check_unate", fallback=True)

    def enable_probabilistic_solver_check_unate(self):
        self.set("PROBABILISTIC_SOLVER", "check_unate", "True")

    def disable_probabilistic_solver_check_unate(self):
        self.set("PROBABILISTIC_SOLVER", "check_unate", "False")


config = NeurolangConfigParser()

config_dirs = [
    os.path.join(sys.prefix, "config"),
    os.path.dirname(os.path.realpath(__file__)),
]
for d in config_dirs:
    config_file = os.path.join(d, "config.ini")
    if os.path.isfile(config_file):
        LOG.info(f"Reading configuration file for Neurolang: {config_file}")
        config.read(config_file)
        LOG.info(f"Read config file with sections {config.sections()}")
        break



import math
import operator

from .exceptions import RelationalAlgebraError
from .expression_walker import (
    PatternWalker,
    add_match
)
from .expressions import (
    Constant,
    FunctionApplication,
    Symbol,
    sure_is_not_pattern,
)
from .relational_algebra import (
    ColumnInt,
    ColumnStr,
    ConcatenateConstantColumn,
    Difference,
    ExtendedProjection,
    FunctionApplicationListMember,
    GroupByAggregation,
    LeftNaturalJoin,
    NameColumns,
    NAryRelationalAlgebraOperation,
    NaturalJoin,
    NumberColumns,
    Product,
    Projection,
    RelationalAlgebraOperation,
    RelationalAlgebraSolver,
    RenameColumn,
    RenameColumns,
    ReplaceNull,
    Selection,
    UnaryRelationalAlgebraOperation,
    Union,
    eq_,
    int2columnint_constant,
    str2columnstr_constant
)
from .utils import OrderedSet

ADD = Constant(operator.add)
MUL = Constant(operator.mul)
SUB = Constant(operator.sub)
LT = Constant(operator.lt)
EQ = Constant(operator.eq)
SUM = Constant(sum)
EXP = Constant(math.exp)
LOG = Constant(math.log)
ONE = Constant[float](1.)
LARGE_NEG = Constant[float](-1e20)


def check_do_not_share_non_prov_col(prov_set_1, prov_set_2):
    shared_columns = set(prov_set_1.non_provenance_columns) & set(
        prov_set_2.non_provenance_columns
    )
    if len(shared_columns) > 0:
        raise RelationalAlgebraError(
            "Provenance sets should not share non-provenance columns. "
            "Shared columns found: {}".format(
                ", ".join(repr(col) for col in shared_columns)
            )
        )


def is_provenance_operation(operation):
    stack = list(operation.unapply())
    while stack:
        stack_element = stack.pop()
        if isinstance(stack_element, ProvenanceAlgebraSet):
            return True
        if isinstance(stack_element, tuple):
            stack += list(stack_element)
        elif isinstance(stack_element, RelationalAlgebraOperation):
            stack += list(stack_element.unapply())
    return False


class ProvenanceAlgebraSet(UnaryRelationalAlgebraOperation):
    def __init__(self, relation, provenance_column):
        self.relation = relation
        self.provenance_column = provenance_column
        self._columns = None
        self._non_provenance_columns = None

    def columns(self):
        if self._columns is None:
            if isinstance(self.relation, Constant):
                columns = OrderedSet(
                    str2columnstr_constant(c) if isinstance(c, str)
                    else int2columnint_constant(c)
                    for c in self.relation.value.columns
                )
            else:
                columns = self.relation.columns()
            self._columns = columns
        return self._columns

    @property
    def non_provenance_columns(self):
        if self._non_provenance_columns is None:
            self._non_provenance_columns = (
                self.columns() - {self.provenance_column}
            )
        return self._non_provenance_columns

    def __repr__(self):
        return (f"RAP[{self.relation}, {self.provenance_column}]")


class NaturalJoinInverse(NaturalJoin):
    def __repr__(self):
        return (
            f"[{self.relation_left}"
            f"INVERSE \N{JOIN}"
            f"{self.relation_right}]"
        )


class WeightedNaturalJoin(NAryRelationalAlgebraOperation):
    def __init__(self, relations, weights):
        self.relations = relations
        self.weights = weights

    def __repr__(self):
        return (
            "\N{Greek Capital Letter Sigma}"
            f"_{self.weights}({self.relations})"
        )


class LiftedPlanProjection(UnaryRelationalAlgebraOperation):
    def __init__(self, relation, attributes):
        self.relation = relation
        self.attributes = attributes

    def columns(self):
        return OrderedSet(self.attributes)


class IndependentProjection(LiftedPlanProjection):
    def __repr__(self) -> str:
        if self.attributes is Ellipsis:
            attributes_repr = "..."
        else:
            attributes_repr = ",".join(repr(attr) for attr in self.attributes)
        return "ind-∃-π_[{}]({})".format(attributes_repr, repr(self.relation))


class IndependentProjectionUniversal(LiftedPlanProjection):
    def __repr__(self) -> str:
        if self.attributes is Ellipsis:
            attributes_repr = "..."
        else:
            attributes_repr = ",".join(repr(attr) for attr in self.attributes)
        return "ind-∀-π_[{}]({})".format(attributes_repr, repr(self.relation))


class DisjointProjection(LiftedPlanProjection):
    def __repr__(self) -> str:
        if self.attributes is Ellipsis:
            attributes_repr = "..."
        else:
            attributes_repr = ",".join(repr(attr) for attr in self.attributes)
        return "disj-π_[{}]({})".format(attributes_repr, repr(self.relation))


class WeightedNaturalJoinSolverMixin(PatternWalker):
    @add_match(
        WeightedNaturalJoin,
        lambda join_op: all(
            isinstance(relation, ProvenanceAlgebraSet)
            for relation in join_op.relations
        )
    )
    def prov_weighted_join(self, join_op):
        relations = join_op.relations
        weights = join_op.weights

        prov_columns = [
            str2columnstr_constant(Symbol.fresh().name)
            for _ in relations
        ]

        dst_columns = relations[0].non_provenance_columns
        for relation in relations[1:]:
            dst_columns = dst_columns | relation.non_provenance_columns
        dst_columns = tuple(dst_columns)

        relations = [
            ExtendedProjection(
                relation.relation,
                (FunctionApplicationListMember(
                    weight * relation.provenance_column,
                    prov_column
                ),) + tuple(
                    FunctionApplicationListMember(
                        c, c
                    )
                    for c in relation.non_provenance_columns
                )
            )
            for relation, weight, prov_column in
            zip(relations, weights, prov_columns)
        ]

        relation = relations[0]
        for relation_ in relations[1:]:
            relation = NaturalJoin(relation, relation_)

        prov_col = str2columnstr_constant(Symbol.fresh().name)
        dst_prov_expr = sum(prov_columns[1:], prov_columns[0])
        relation = ExtendedProjection(
            relation,
            (FunctionApplicationListMember(
                dst_prov_expr,
                prov_col
            ),) + tuple(
                FunctionApplicationListMember(
                    c, c
                )
                for c in dst_columns
            )
        )

        return self.walk(ProvenanceAlgebraSet(
            relation,
            prov_col
        ))


class IndependentDisjointProjectionsAndUnionMixin(PatternWalker):
    @add_match(IndependentProjection(ProvenanceAlgebraSet, ...))
    def independent_projection(self, proj_op):
        prov_set = proj_op.relation
        prov_col = prov_set.provenance_column
        proj_list = [
            FunctionApplicationListMember(col, col)
            for col in prov_set.non_provenance_columns
        ]
        proj_list_log = proj_list + [
            FunctionApplicationListMember(
                LOG(ONE - prov_col),
                prov_col,
            )
        ]
        proj_list_0 = proj_list + [
            FunctionApplicationListMember(
                LARGE_NEG,
                prov_col,
            )
        ]
        relation_log = ExtendedProjection(
            Selection(prov_set.relation, FunctionApplication(LT, (prov_col, ONE))),
            proj_list_log
        )
        relation_zero = ExtendedProjection(
            Selection(prov_set.relation, FunctionApplication(EQ, (prov_col, ONE))),
            proj_list_0
        )
        relation = Union(relation_log, relation_zero)
        relation = GroupByAggregation(
            relation,
            groupby=proj_op.attributes,
            aggregate_functions=(
                FunctionApplicationListMember(
                    FunctionApplication(Constant(sum), (prov_col,)),
                    prov_col,
                ),
            ),
        )
        proj_list = [
            FunctionApplicationListMember(col, col)
            for col in relation.groupby
        ]
        proj_list.append(
            FunctionApplicationListMember(
               ONE - EXP(prov_col),
               prov_col,
            )
        )

        relation = ExtendedProjection(relation, proj_list)
        return ProvenanceAlgebraSet(relation, prov_col)

    @add_match(IndependentProjectionUniversal(ProvenanceAlgebraSet, ...))
    def independent_projection_universal(self, proj_op):
        prov_set = proj_op.relation
        prov_col = prov_set.provenance_column
        proj_list = [
            FunctionApplicationListMember(col, col)
            for col in prov_set.non_provenance_columns
        ]
        proj_list.append(
            FunctionApplicationListMember(
                LOG(prov_col),
                prov_col,
            )
        )
        relation = ExtendedProjection(prov_set.relation, proj_list)
        relation = GroupByAggregation(
            relation,
            groupby=proj_op.attributes,
            aggregate_functions=(
                FunctionApplicationListMember(
                    FunctionApplication(Constant(sum), (prov_col,)),
                    prov_col,
                ),
            ),
        )
        proj_list = [
            FunctionApplicationListMember(col, col)
            for col in relation.groupby
        ]
        proj_list.append(
            FunctionApplicationListMember(
               EXP(prov_col),
               prov_col,
            )
        )
        relation = ExtendedProjection(relation, proj_list)
        return ProvenanceAlgebraSet(relation, prov_col)

    @add_match(DisjointProjection(ProvenanceAlgebraSet, ...))
    def disjoint_projection(self, proj_op):
        prov_set = proj_op.relation
        prov_col = prov_set.provenance_column
        aggregate_functions = [
            FunctionApplicationListMember(
                FunctionApplication(Constant(sum), (prov_col,)),
                prov_col,
            ),
        ]
        operation = GroupByAggregation(
            prov_set.relation,
            proj_op.attributes,
            aggregate_functions,
        )
        res = ProvenanceAlgebraSet(operation, prov_col)
        return res

    @add_match(Union(ProvenanceAlgebraSet, ProvenanceAlgebraSet))
    def union_rap(self, union):
        prov_column_left = union.relation_left.provenance_column
        prov_column_right = union.relation_right.provenance_column
        relation_left = union.relation_left.relation
        relation_right = union.relation_right.relation
        if prov_column_left != prov_column_right:
            relation_right = RenameColumn(
                relation_right,
                prov_column_right,
                prov_column_left,
            )
        columns_to_keep = union.relation_left.non_provenance_columns
        operation = IndependentProjection(
            ProvenanceAlgebraSet(
                Union(relation_left, relation_right),
                prov_column_left
            ),
            columns_to_keep
        )
        return self.walk(operation)


class BuildProvenanceAlgebraSetWalkIntoMixin(PatternWalker):
    @add_match(
        ProvenanceAlgebraSet,
        lambda exp: any(not isinstance(arg, Constant) for arg in exp.unapply())
    )
    def cycle_in_build_provenance_algebra_set(self, expression):
        relation = self.walk(expression.relation)
        provenance_column = self.walk(expression.provenance_column)
        if (
            (relation is not expression.relation) or
            (provenance_column is not expression.provenance_column)
        ):
            return self.walk(
                ProvenanceAlgebraSet(relation, provenance_column)
            )
        else:
            return expression


class ProvenanceSelectionMixin(PatternWalker):
    @add_match(
        Selection(
            ProvenanceAlgebraSet,
            FunctionApplication(
                eq_, (Constant[ColumnInt], Constant[ColumnInt])
            )
        )
    )
    def selection_rap_eq_columnint_columnint(self, selection):
        columns = selection.relation.non_provenance_columns
        formula = selection.formula
        new_formula = FunctionApplication(
            eq_, (
                columns[formula.args[0].value],
                columns[formula.args[1].value],
            )
        )
        return self.walk(
            ProvenanceAlgebraSet(
                Selection(selection.relation.relation, new_formula),
                selection.relation.provenance_column
            )
        )

    @add_match(
        Selection(
            ProvenanceAlgebraSet,
            FunctionApplication(eq_, (Constant[ColumnInt], ...))
        )
    )
    def selection_rap_eq_columnint(self, selection):
        columns = selection.relation.non_provenance_columns
        formula = selection.formula
        new_formula = FunctionApplication(
            eq_, (
                columns[formula.args[0].value],
                formula.args[1]
            )
        )
        new_relation = ProvenanceAlgebraSet(
            Selection(selection.relation.relation, new_formula),
            selection.relation.provenance_column
        )
        return self.walk(new_relation)

    @add_match(
        Selection(ProvenanceAlgebraSet, ...)
    )
    def selection_provenance_set(self, selection):
        res = ProvenanceAlgebraSet(
            Selection(
                selection.relation.relation,
                selection.formula
            ),
            selection.relation.provenance_column
        )
        return self.walk(res)


class ProvenanceColumnManipulationMixin(PatternWalker):
    @add_match(ConcatenateConstantColumn(ProvenanceAlgebraSet, ..., ...))
    def prov_concatenate_constant_column(self, concat_op):
        if concat_op.column_name == concat_op.relation.provenance_column:
            new_prov_col = str2columnstr_constant(Symbol.fresh().name)
            relation = RenameColumn(
                concat_op.relation.relation,
                concat_op.relation.provenance_column,
                new_prov_col,
            )
        else:
            relation = concat_op.relation.relation
            new_prov_col = concat_op.relation.provenance_column

        res = ProvenanceAlgebraSet(
            ConcatenateConstantColumn(
                relation,
                concat_op.column_name,
                concat_op.column_value
            ),
            new_prov_col
        )
        return self.walk(res)

    @add_match(
        NameColumns(ProvenanceAlgebraSet, ...),
        lambda expression: all(
            isinstance(column, Constant[ColumnInt])
            for column in expression.relation.non_provenance_columns
        )
    )
    def name_columns_int_rap(self, expression):
        relation = expression.relation.relation
        rap_column = str2columnstr_constant(Symbol.fresh().name)
        rap_column_index = expression.relation.provenance_column.value
        new_columns_named = (
            expression.column_names[:rap_column_index] +
            (rap_column,) +
            expression.column_names[rap_column_index:]
        )
        new_relation = ProvenanceAlgebraSet(
            NameColumns(
                relation,
                new_columns_named
            ),
            new_columns_named[0]
        )

        return self.walk(new_relation)

    @add_match(NameColumns(ProvenanceAlgebraSet, ...))
    def name_columns_rap(self, expression):
        relation = expression.relation.relation
        columns_to_name = (
            c for c in expression.relation.columns()
            if c != expression.relation.provenance_column
        )
        ne = RenameColumns(
            relation,
            tuple(
                (src, dst)
                for src, dst in zip(
                    columns_to_name, expression.column_names
                )
            )
        )
        return self.walk(ProvenanceAlgebraSet(
            ne,
            expression.relation.provenance_column
        ))

    @add_match(NumberColumns(ProvenanceAlgebraSet, ...))
    def number_columns_rap(self, expression):
        new_columns = (
            (expression.relation.provenance_column,) +
            expression.column_names
        )
        new_expression = ProvenanceAlgebraSet(
            NumberColumns(
                expression.relation.relation,
                new_columns
            ),
            int2columnint_constant(0)
        )
        return self.walk(new_expression)

    @add_match(RenameColumn(ProvenanceAlgebraSet, ..., ...))
    def prov_rename_column(self, rename_column):
        if rename_column.src == rename_column.relation.provenance_column:
            provenance_column = rename_column.dst
        else:
            provenance_column = rename_column.relation.provenance_column
        res = ProvenanceAlgebraSet(
            RenameColumn(
                rename_column.relation.relation,
                rename_column.src,
                rename_column.dst
            ),
            provenance_column
        )
        return self.walk(res)

    @add_match(RenameColumns(ProvenanceAlgebraSet, ...))
    def prov_rename_columns(self, rename_columns):
        res = ProvenanceAlgebraSet(
            RenameColumns(
                rename_columns.relation.relation,
                rename_columns.renames
            ),
            rename_columns.relation.provenance_column
        )
        return self.walk(res)


class ProvenanceExtendedProjectionMixin(PatternWalker):
    """
    Mixin that implements specific cases of extended projections on provenance
    sets for which the semantics are not modified.

    An extended projection on a provenance set is allowed if all the
    non-provenance columns `c` are projected _as is_ (i.e. `c -> c`), which
    ensures the number of tuples is going to be exactly the same in the
    resulting provenance set, and the provenance label is still going to be
    semantically valid. This is useful in particular for projecting a constant
    column, which can happen when dealing with rules with constant terms or
    variable equalities.

    """
    @add_match(ExtendedProjection(ProvenanceAlgebraSet, ...))
    def prov_extended_projection(self, extended_proj):
        relation = extended_proj.relation
        equality_columns = {
            c.dst_column for c in extended_proj.projection_list
            if c.fun_exp == c.dst_column
        }
        if equality_columns != relation.non_provenance_columns:
            raise ValueError("Invalid provenance ExtendedProjection")

        if any(
            proj_list_member.dst_column == relation.provenance_column
            for proj_list_member in extended_proj.projection_list
        ):
            new_prov_col = str2columnstr_constant(Symbol.fresh().name)
            relation = RenameColumn(
                relation,
                relation.provenance_column,
                new_prov_col
            )
        else:
            new_prov_col = relation.provenance_column

        new_proj_list = extended_proj.projection_list + (
            FunctionApplicationListMember(
                fun_exp=new_prov_col, dst_column=new_prov_col
            ),
        )
        return self.walk(ProvenanceAlgebraSet(
                ExtendedProjection(
                    relation.relation, new_proj_list,
                ),
                new_prov_col
        ))


class ProvenanceSetOperationsMixin(PatternWalker):
    @add_match(Union(ProvenanceAlgebraSet, ProvenanceAlgebraSet))
    def prov_union(self, union_op):
        left = union_op.relation_left
        right = union_op.relation_right
        if left.non_provenance_columns != right.non_provenance_columns:
            raise RelationalAlgebraError(
                "All non-provenance columns must be the same: {} != {}".format(
                    left.non_provenance_columns, right.non_provenance_columns,
                )
            )
        prov_col = left.provenance_column
        result_columns = left.non_provenance_columns
        # move to non-provenance relational algebra
        np_left = left.relation
        np_right = right.relation
        # make the the provenance columns match
        np_right = RenameColumn(
            np_right,
            right.provenance_column,
            prov_col,
        )
        # add a dummy column with different values for each relation
        # this ensures that all the tuples will be part of the result
        dummy_col = str2columnstr_constant(Symbol.fresh().name)
        np_left = ConcatenateConstantColumn(
            np_left, dummy_col, Constant[int](0)
        )
        np_right = ConcatenateConstantColumn(
            np_right, dummy_col, Constant[int](1)
        )
        ra_union_op = Union(np_left, np_right)
        result = ProvenanceAlgebraSet(ra_union_op, prov_col)
        # provenance projection that removes the dummy column and sums the
        # provenance of matching tuples
        result = Projection(result, result_columns)
        return self.walk(result)


class RelationalAlgebraProvenanceExpressionSemringSolverMixin(
    ProvenanceSelectionMixin,
    ProvenanceExtendedProjectionMixin,
    ProvenanceColumnManipulationMixin,
    ProvenanceSetOperationsMixin,
):
    @add_match(NaturalJoin(
        ProvenanceAlgebraSet, ProvenanceAlgebraSet
    ))
    def prov_naturaljoin(self, naturaljoin):
        return self.walk(self._apply_provenance_join_operation(
            naturaljoin.relation_left,
            naturaljoin.relation_right,
            NaturalJoin,
            self._semiring_mul,
        ))

    def _apply_provenance_join_operation(
        self, left, right, np_op, prov_binary_op
    ):
        res_columns = set(left.columns()) | (
            set(right.columns()) - {right.provenance_column}
        )
        res_columns = tuple(col for col in res_columns)
        res_prov_col = left.provenance_column
        # provenance columns are temporarily renamed for executing the
        # non-provenance operation on the relations
        tmp_left_col = str2columnstr_constant(
            f"{left.provenance_column.value}1"
        )
        tmp_right_col = str2columnstr_constant(
            f"{right.provenance_column.value}2"
        )
        tmp_left = RenameColumn(
            left.relation,
            left.provenance_column,
            tmp_left_col,
        )
        tmp_right = RenameColumn(
            right.relation,
            right.provenance_column,
            tmp_right_col,
        )
        tmp_np_op_args = (tmp_left, tmp_right)
        if np_op is Product:
            tmp_non_prov_result = np_op(tmp_np_op_args)
        elif np_op is NaturalJoin:
            tmp_non_prov_result = np_op(*tmp_np_op_args)
        else:
            raise RelationalAlgebraError(
                "Cannot apply non-provenance operation: {}".format(np_op)
            )
        result = ExtendedProjection(
            tmp_non_prov_result,
            (
                FunctionApplicationListMember(
                    fun_exp=prov_binary_op(tmp_left_col, tmp_right_col),
                    dst_column=res_prov_col,
                ),
            )
            + tuple(
                FunctionApplicationListMember(fun_exp=col, dst_column=col)
                for col in set(res_columns) - {res_prov_col}
            ),
        )
        return ProvenanceAlgebraSet(
            result, res_prov_col
        )

    @add_match(
        Product,
        lambda product: all(
            isinstance(ras, ProvenanceAlgebraSet)
            for ras in product.relations
        )
    )
    def prov_product(self, product):
        rel_res = product.relations[0]
        for relation in product.relations[1:]:
            check_do_not_share_non_prov_col(rel_res, relation)
            rel_res = self._apply_provenance_join_operation(
                rel_res, relation, Product, self._semiring_mul,
            )
        return self.walk(rel_res)

    def _semiring_mul(self, left, right):
        return left * right

    @add_match(
        Projection(ProvenanceAlgebraSet, ...),
        lambda proj: (
            all(
                issubclass(att.type, ColumnInt) for att in proj.attributes
            ) and issubclass(proj.relation.provenance_column.type, ColumnInt)
            and (
                len(proj.attributes) ==
                len(proj.relation.non_provenance_columns)
            )
        )
    )
    def projection_rap_columnint(self, projection):
        columns = projection.relation.non_provenance_columns
        new_attributes = (projection.relation.provenance_column,)
        for att in projection.attributes:
            if issubclass(att.type, ColumnInt):
                att = columns[att.value]
            new_attributes += (att,)
        return self.walk(
            ProvenanceAlgebraSet(
                Projection(
                    projection.relation.relation,
                    new_attributes
                ),
                int2columnint_constant(0)
            )
        )

    @add_match(
        Projection(ProvenanceAlgebraSet, ...),
        lambda proj: (
            all(
                issubclass(att.type, ColumnInt) for att in proj.attributes
            ) and issubclass(proj.relation.provenance_column.type, ColumnInt)
        )
    )
    def projection_rap_columnint_subset(self, projection):
        columns = projection.relation.non_provenance_columns
        new_columns = tuple(str2columnstr_constant(f'col_{i}') for i in range(len(columns)))
        new_attributes = tuple(
            new_columns[a.value] for a in projection.attributes
        )
        new_expression = NumberColumns(
            Projection(
                NameColumns(
                    projection.relation,
                    new_columns
                ),
                new_attributes
            ),
            new_attributes
        )
        return self.walk(new_expression)

    @add_match(
        Projection(ProvenanceAlgebraSet, ...),
        lambda proj: all(
            issubclass(att.type, ColumnStr) for att in proj.attributes
        )
    )
    def projection_rap(self, projection):
        if (
            set(projection.attributes) ==
            projection.relation.non_provenance_columns
        ):
            return self.walk(projection.relation)

        aggregate_functions = (
            FunctionApplicationListMember(
                self._semiring_agg_sum(
                    (projection.relation.provenance_column,)
                ),
                projection.relation.provenance_column
            ),
        )
        operation = GroupByAggregation(
            projection.relation.relation,
            projection.attributes,
            aggregate_functions,
        )

        with sure_is_not_pattern():
            res = self.walk(
                ProvenanceAlgebraSet(
                    operation,
                    projection.relation.provenance_column
                )
            )
        return res

    def _semiring_agg_sum(self, args):
        return FunctionApplication(
            SUM, args, validate_arguments=False, verify_type=False
        )

    @add_match(
        Difference(ProvenanceAlgebraSet, ProvenanceAlgebraSet)
    )
    def difference(self, diff):
        left = diff.relation_left
        right = diff.relation_right
        res_columns = tuple(
            col for col in left.columns()
        )
        res_prov_col = left.provenance_column
        tmp_left_prov_col = str2columnstr_constant(Symbol.fresh().name)
        tmp_right_prov_col = str2columnstr_constant(Symbol.fresh().name)
        tmp_left = RenameColumn(
            left.relation,
            left.provenance_column,
            tmp_left_prov_col,
        )
        tmp_right = RenameColumn(
            right.relation,
            right.provenance_column,
            tmp_right_prov_col,
        )
        tmp_np_op_args = (tmp_left, tmp_right)
        tmp_non_prov_result = LeftNaturalJoin(*tmp_np_op_args)
        tmp_non_prov_result = ReplaceNull(
            tmp_non_prov_result,
            tmp_right_prov_col,
            Constant(0.)
        )
        result = ExtendedProjection(
            tmp_non_prov_result,
            (
                FunctionApplicationListMember(
                    fun_exp=self._semiring_monus(
                        tmp_left_prov_col, tmp_right_prov_col
                    ),
                    dst_column=res_prov_col,
                ),
            )
            + tuple(
                FunctionApplicationListMember(fun_exp=col, dst_column=col)
                for col in set(res_columns) - {res_prov_col}
            ),
        )
        return self.walk(ProvenanceAlgebraSet(
            result, res_prov_col
        ))

    def _semiring_monus(self, left, right):
        return MUL(
            left,
            SUB(Constant(1), right)
        )


class RelationalAlgebraProvenanceCountingSolverMixin(
    RelationalAlgebraProvenanceExpressionSemringSolverMixin,
    PatternWalker
):
    """
    Mixing that walks through relational algebra expressions and
    executes the operations and provenance calculations.

    """

    @add_match(NaturalJoinInverse)
    def prov_naturaljoin_inverse(self, naturaljoin):
        return self.walk(self._apply_provenance_join_operation(
            naturaljoin.relation_left,
            naturaljoin.relation_right,
            NaturalJoin,
            Constant(operator.truediv),
        ))


class RelationalAlgebraProvenanceCountingSolver(
    BuildProvenanceAlgebraSetWalkIntoMixin,
    RelationalAlgebraProvenanceCountingSolverMixin,
    RelationalAlgebraSolver,
):
    pass


class RelationalAlgebraProvenanceExpressionSemringSolver(
    BuildProvenanceAlgebraSetWalkIntoMixin,
    RelationalAlgebraProvenanceExpressionSemringSolverMixin,
    RelationalAlgebraSolver,
):
    pass
